// jl: Display-Listen und Vertex-Arrays

#include "stdafx.h"
#include "resource.h"

#define MAX_LOADSTRING 100

const double pi = 4.0*atan(1.0);

HINSTANCE hInst; // aktuelle Instanz
TCHAR szTitle[MAX_LOADSTRING]; // Text der Titelzeile
TCHAR szWindowClass[MAX_LOADSTRING]; // Text der Titelzeile

ATOM				MyRegisterClass( HINSTANCE hInstance );
BOOL				InitInstance( HINSTANCE, int );
LRESULT CALLBACK	WndProc( HWND, UINT, WPARAM, LPARAM );
LRESULT CALLBACK	About( HWND, UINT, WPARAM, LPARAM );

static unsigned char* myTexture;
static int texWidth, texHeight;

double frac(double x)
{
	return x-(int)x;
}

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow )
{
	MSG msg;
	HACCEL hAccelTable;

	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
	LoadString(hInstance, IDC_OGLRUMPFPROGRAMM, szWindowClass, MAX_LOADSTRING);
	MyRegisterClass(hInstance);

	// jl
	// RGB-Textur laden
	// Muss hier 24-Bit-BMP-Ressource sein.
	// Bildbreite und -hhe mssen Zweierpotenzen und Vielfache von 4 sein.

	HRSRC bmpRes = FindResource(hInstance, MAKEINTRESOURCE(IDB_RGB), RT_BITMAP);
	if(!bmpRes)
	{
		return FALSE;
	}
	HGLOBAL hBmpRes = LoadResource(hInstance, bmpRes);
	if(!hBmpRes)
	{
		return FALSE;
	}
	BITMAPINFO* pBmp = (BITMAPINFO*)LockResource(hBmpRes);
	if(!pBmp)
	{
		return FALSE;
	}
	texHeight = pBmp->bmiHeader.biHeight;
	texWidth = pBmp->bmiHeader.biWidth;
	// Prfe auf korrekte Abmessungen
	if(texHeight%4 != 0
		|| texWidth%4 != 0
		|| fabs(frac(log(texHeight)/log(2)))>1e-10
		|| fabs(frac(log(texWidth)/log(2)))>1e-10
		|| pBmp->bmiHeader.biBitCount != 24)
	{
		return FALSE;
	}

	myTexture = new unsigned char[4*texHeight*texWidth];

	int i;
	for(i=0; i<texHeight*texWidth; ++i)
	{
		myTexture[4*i] =
			(((unsigned char*)pBmp) + pBmp->bmiHeader.biSize)[3*i];
		myTexture[4*i+1] =
			(((unsigned char*)pBmp) + pBmp->bmiHeader.biSize)[3*i+1];
		myTexture[4*i+2] =
			(((unsigned char*)pBmp) + pBmp->bmiHeader.biSize)[3*i+2];
		myTexture[4*i+3] = 255; // Alpha nicht aus Datei
	}

	DeleteObject(hBmpRes);	
	
	// Initialisierung der Anwendung durchfhren:
	if( !InitInstance( hInstance, nCmdShow ) ) 
	{
		return FALSE;
	}

	hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_OGLRUMPFPROGRAMM);

	// Hauptnachrichtenschleife:
	while( GetMessage(&msg, NULL, 0, 0) ) 
	{
		if( !TranslateAccelerator (msg.hwnd, hAccelTable, &msg) ) 
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
	}

	delete[] myTexture;

	return msg.wParam;
}

//
//  FUNKTION: MyRegisterClass()
//
//  AUFGABE: Registriert die Fensterklasse.
ATOM MyRegisterClass( HINSTANCE hInstance )
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX); 

	wcex.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_OGLRUMPFPROGRAMM);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)NULL;
	wcex.lpszMenuName	= (LPCSTR)IDC_OGLRUMPFPROGRAMM;
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);

	return RegisterClassEx(&wcex);
}

//
//   FUNKTION: InitInstance(HANDLE, int)
//
//   AUFGABE: Speichert die Instanzzugriffsnummer und erstellt das Hauptfenster
BOOL InitInstance( HINSTANCE hInstance, int nCmdShow )
{
   HWND hWnd;

   hInst = hInstance; // Instanzzugriffsnummer in unserer globalen Variable speichern

   hWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

   if( !hWnd ) 
   {
      return FALSE;
   }

   ShowWindow( hWnd, nCmdShow );
   UpdateWindow( hWnd );

   return TRUE;
}

//  FUNKTION: WndProc(HWND, unsigned, WORD, LONG)
//
//  AUFGABE:  Verarbeitet Nachrichten fr das Hauptfenster.
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;
	static HDC hDC = NULL;
	static HGLRC hRC = NULL;

	static double angleH = 13.0;
	static double angleV = 15.0;

	static GLuint displayList1;

	static GLfloat* vertices ;
	static GLfloat* normals;
	static GLfloat* texcoords;
	static GLuint* indices;
	static int nVertices;

	switch( message ) 
	{
		case WM_CREATE:
			{
				hDC = GetDC(hWnd);
				static PIXELFORMATDESCRIPTOR pfd = {
					sizeof(PIXELFORMATDESCRIPTOR),
					1,
					PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
					PFD_TYPE_RGBA,
					32,
					0,0,0,0,0,0,0,0,0,0,0,0,0,
					16, // Bit-Tiefe des z-Buffer
					0,0,0,0,0,0,0 };
				int nPixelFormat = ChoosePixelFormat(hDC, &pfd);
				if(nPixelFormat == 0)
				{
					MessageBox(hWnd, "Initialisierung von OpenGL fehlgeschlagen.", "Fataler Fehler", MB_ICONERROR);
					PostQuitMessage(0);
				} else {
					SetPixelFormat(hDC, nPixelFormat, &pfd);
					hRC = wglCreateContext(hDC);
					
					wglMakeCurrent(hDC, hRC); // bleibt hier "current" bis Programmende

					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
					glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					gluBuild2DMipmaps(GL_TEXTURE_2D, 4,
						texWidth, texHeight, GL_RGBA, GL_UNSIGNED_BYTE,
						myTexture);

					GLfloat lightPos[] = {-10.0f, 10.0f, 3.0f, 1.0f};
					glLightfv(GL_LIGHT0, GL_POSITION, lightPos);

					glEnable(GL_LIGHT0);
					GLfloat diffCol[] = {1.0f, 1.0f, 0.4f, 1.0f};
					glLightfv(GL_LIGHT0, GL_DIFFUSE, diffCol);
					GLfloat specCol[] = {1.0f, 1.0f, 1.0f, 1.0f};
					glLightfv(GL_LIGHT0, GL_SPECULAR, specCol);
					GLfloat ambCol[] = {0.2f, 0.2f, 0.2f, 1.0f};
					glLightfv(GL_LIGHT0, GL_AMBIENT, ambCol);

					GLfloat matDiffCol[] = {1.0f, 1.0f, 1.0f, 1.0f};
					glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, matDiffCol);
					GLfloat matSpecCol[] = {1.0f, 1.0f, 1.0f, 1.0f};
					glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, matSpecCol);
					glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 256.0);
					GLfloat matAmbCol[] = {1.0f, 1.0f, 1.0f, 1.0f};
					glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, matAmbCol);
					GLfloat matEmCol[] = {0.1f, 0.0f, 0.0f, 1.0f};
					glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, matEmCol);

					glEnable(GL_DEPTH_TEST);
					glEnable(GL_CULL_FACE);
					glEnable(GL_LIGHTING);
					glEnable(GL_TEXTURE_2D);

					//****Beispiel fr eine Display-Liste****
					// Kennummer vom Treiber geben lassen
					displayList1 = glGenLists(1);
					// Aufzeichnung starten
					glNewList(displayList1, GL_COMPILE);

					glBegin(GL_QUADS);
						glNormal3d(0.0, 1.0, 0.0);
						glTexCoord2d(0.0, 0.0);
						glVertex3d(-2.0, 0.0,  2.0);
						glTexCoord2d(1.0, 0.0);
						glVertex3d( 2.0, 0.0,  2.0);
						glTexCoord2d(1.0, 1.0);
						glVertex3d( 2.0, 0.0, -2.0);
						glTexCoord2d(0.0, 1.0);
						glVertex3d(-2.0, 0.0, -2.0);

						glNormal3d(0.0, -1.0, 0.0);
						glTexCoord2d(0.1, 0.1);
						glVertex3d(-2.0, -0.01,  2.0);
						glTexCoord2d(0.3, 0.1);
						glVertex3d(-2.0, -0.01, -2.0);
						glTexCoord2d(0.3, 0.3);
						glVertex3d( 2.0, -0.01, -2.0);
						glTexCoord2d(0.1, 0.3);
						glVertex3d( 2.0, -0.01,  2.0);
					glEnd();

					// Aufzeichnung beenden
					glEndList();

					//****(indizierten) Vertex-Array initialisieren****
					// Geometrie: die Punkte, Normalen und Texturkoordinaten
					const int stacks = 120;
					const int slices = 80;
					vertices = new GLfloat[3*(stacks+1)*(slices+1)];
					normals = new GLfloat[3*(stacks+1)*(slices+1)];
					texcoords = new GLfloat[2*(stacks+1)*(slices+1)];
					int i, j;
					for(i = 0; i<=stacks; ++i)
					{
						double theta = pi*i/(double)stacks;
						for(j = 0; j<=slices; ++j)
						{
							double phi = 2.0*pi*j/(double)slices;
							vertices[(i*(slices+1)+j)*3  ] = (float)(sin(theta)*cos(phi));
							vertices[(i*(slices+1)+j)*3+1] = (float)(sin(theta)*sin(phi));
							vertices[(i*(slices+1)+j)*3+2] = (float)(cos(theta));

							// Die Normalen sind hier nur zufllig gleich den Ortsvektoren.
							normals[(i*(slices+1)+j)*3  ] = (float)(sin(theta)*cos(phi));
							normals[(i*(slices+1)+j)*3+1] = (float)(sin(theta)*sin(phi));
							normals[(i*(slices+1)+j)*3+2] = (float)(cos(theta));

							texcoords[(i*(slices+1)+j)*2  ] = (float)(j/(double)slices);
							texcoords[(i*(slices+1)+j)*2+1] = (float)(1.0-i/(double)stacks);
						}
					}
					// Topologie: wie die Vertizes zusammenhngen
					nVertices = 4*stacks*slices;
					indices = new GLuint[nVertices];
					int c = 0;
					for(i = 0; i<stacks; ++i)
					{
						for(j = 0; j<slices; ++j)
						{
							indices[c++] = i*(slices+1)+j;
							indices[c++] = (i+1)*(slices+1)+j;
							indices[c++] = (i+1)*(slices+1)+(j+1);
							indices[c++] = i*(slices+1)+(j+1);
						}
					}
				}
			}
			break;
		case WM_SIZE:
			{
				wglMakeCurrent(hDC, hRC);
				RECT r;
				GetClientRect(hWnd, &r);
				glViewport(r.left, r.top, r.right-r.left, r.bottom-r.top);

				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				int w = r.bottom-r.top;
				if(w == 0)
				{
					w = 1; // gegen Division durch null im nchsten Schritt
				}
				gluPerspective(40.0, (r.right-r.left)/(double)w, 1.0, 40.0);
			}
			break;
		case WM_COMMAND:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			switch( wmId ) 
			{
				case IDM_ABOUT:
				   DialogBox(hInst, (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
				   break;
				case IDM_EXIT:
				   DestroyWindow( hWnd );
				   break;
				default:
				   return DefWindowProc( hWnd, message, wParam, lParam );
			}
			break;
		case WM_PAINT:
			{
				glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();

				glTranslated(0.0, 0.0, -5.0);
				glRotated(angleH, 0.0, 1.0, 0.0);
				glRotated(angleV, 1.0, 0.0, 0.0);

				// Display-Liste ausfhren
				glCallList(displayList1);

				// Vertex-Array aufrufen
				glEnableClientState(GL_VERTEX_ARRAY);
				glVertexPointer(3, GL_FLOAT, 0, vertices);
				glEnableClientState(GL_NORMAL_ARRAY);
				glNormalPointer(GL_FLOAT, 0, normals);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, 0, texcoords);

				glDrawElements(GL_QUADS, nVertices, GL_UNSIGNED_INT, indices);

				glDisableClientState(GL_VERTEX_ARRAY);
				glDisableClientState(GL_NORMAL_ARRAY);
				glDisableClientState(GL_TEXTURE_COORD_ARRAY);

				SwapBuffers(hDC);

				ValidateRect(hWnd, NULL);
			}
			break;
		case WM_CHAR:
			switch(wParam)
			{
				case 'h':
					angleH += 5.0;
					InvalidateRect(hWnd, NULL, FALSE);
					break;
				case 'v':
					angleV += 5.0;
					InvalidateRect(hWnd, NULL, FALSE);
					break;
				default: ;
			}
			break;
		case WM_DESTROY:
			glDeleteLists(displayList1, 1);

			delete[] vertices;
			delete[] normals;
			delete[] texcoords;
			delete[] indices;

			wglMakeCurrent(hDC, NULL);
			wglDeleteContext(hRC);
			PostQuitMessage( 0 );
			break;
		default:
			return DefWindowProc( hWnd, message, wParam, lParam );
   }
   return 0;
}

// Nachrichtenbehandlungsroutine fr "Info"-Feld.
LRESULT CALLBACK About( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam )
{
	switch( message )
	{
		case WM_INITDIALOG:
				return TRUE;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
			{
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}
    return FALSE;
}